from typing import List, Optional

import arrow
from tinydb import Query, TinyDB
from tinydb.operations import set
from tinydb.table import Document

from .config import config
from .result import Result


def exist_in_cache(db: TinyDB, image_md5: str, mode: str) -> Optional[Result]:
    result: Optional[Result] = None
    cache_result: List[Document] = db.search(
        (Query().image_md5 == image_md5) & (Query().mode == mode)
    )
    if cache_result:
        db.update(
            set("update_at", arrow.now().for_json()), Query().image_md5 == image_md5  # type: ignore
        )
        result = Result(cache_result[-1])
    return result


def clear_expired_cache(db: TinyDB) -> None:
    expired_date = arrow.now().shift(days=-config.cache_expire).for_json()
    db.remove(Query().update_at < expired_date)



from typing import Optional

from nonebot import get_driver
from pydantic import BaseSettings


class Config(BaseSettings):

    hide_img: bool = False  # 隐藏所有搜索结果的缩略图
    saucenao_low_acc: int = 60  # saucenao 相似度低于这个百分比将被认定为相似度过低
    use_ascii2d_when_low_acc: bool = True  # 是否在 saucenao 相似度过低时自动使用 ascii2d
    group_forward_search_result: bool = True  # 若结果消息有多条，采用合并转发方式发送搜图结果（私聊不生效）
    proxy: Optional[str] = None  # 大部分请求所使用的代理: http(s)://
    cache_expire: int = 7  # 搜图结果缓存过期时间（天）
    saucenao_api_key: str = ""  # saucenao APIKEY，必填，否则无法使用 saucenao 搜图
    exhentai_cookies: str = ""  # exhentai cookies，选填，没有的情况下自动改用 e-hentai 搜图

    class Config:
        extra = "ignore"


global_config = get_driver().config
config = Config(**global_config.dict())



import math
from typing import Any, Dict, List, Optional

from PicImageSearch import Network, TraceMoe

from .utils import handle_img


async def whatanime_search(url: str, proxy: Optional[str], hide_img: bool) -> List[str]:
    async with Network(proxies=proxy) as client:
        whatanime = TraceMoe(client=client)
        res = await whatanime.search(url=url)
        if res and res.raw:
            time = res.raw[0].From
            minutes = math.floor(time / 60)
            seconds = math.floor(time % 60)
            time_str = f"{minutes:02d}:{seconds:02d}"
            cover_image = await handle_img(
                res.raw[0].anime_info["coverImage"]["large"],
                proxy,
                hide_img,
            )
            chinese_title = res.raw[0].title_chinese
            native_title = res.raw[0].title_native
            _type = res.raw[0].anime_info["type"]
            _format = res.raw[0].anime_info["format"]

            def date_to_str(date: Dict[str, Any]) -> str:
                return f"{date['year']}-{date['month']}-{date['day']}"

            start_date = date_to_str(res.raw[0].anime_info["startDate"])
            end_date = ""
            if res.raw[0].anime_info["endDate"]["year"] > 0:
                end_date = date_to_str(res.raw[0].anime_info["endDate"])
            res_list = [
                f"WhatAnime（{res.raw[0].similarity}%）",
                f"该截图出自第 {res.raw[0].episode} 集的 {time_str}",
                cover_image,
                chinese_title,
                native_title,
                f"类型：{_type}-{_format}",
                f"开播：{start_date}",
                f"完结：{end_date}" if end_date else "",
            ]
            return ["\n".join([i for i in res_list if i != ""])]
        return ["WhatAnime 暂时无法使用"]



from typing import List, Optional, Union

import aiohttp
from lxml.html import HTMLParser, fromstring
from PicImageSearch import EHentai, Network
from PicImageSearch.model import EHentaiItem, EHentaiResponse
from pyquery import PyQuery

from .config import config
from .utils import handle_img


class EHentaiResponseAioHttp:
    def __init__(self, resp_text: str, resp_url: str):
        self.origin: str = resp_text
        utf8_parser = HTMLParser(encoding="utf-8")
        data = PyQuery(fromstring(self.origin, parser=utf8_parser))
        self.raw: List[EHentaiItem] = [
            EHentaiItem(i) for i in data.find(".glcat").parents("tr").items()
        ]
        self.url: str = resp_url


async def ehentai_search(url: str, proxy: Optional[str], hide_img: bool) -> List[str]:
    async with Network(proxies=proxy, cookies=config.exhentai_cookies) as client:
        ex = bool(config.exhentai_cookies)
        ehentai = EHentai(client=client)
        if res := await ehentai.search(url, ex=ex):
            if not res.raw:
                # 如果第一次没找到，使搜索结果包含被删除的部分，并重新搜索
                ehentai = EHentai(client=client, exp=True)
                res = await ehentai.search(url, ex=ex)
            return await search_result_filter(res, proxy, hide_img)
        return ["EHentai 暂时无法使用"]


async def ehentai_title_search(
    title: str, proxy: Optional[str], hide_img: bool
) -> List[str]:
    headers = {}
    if cookies := config.exhentai_cookies:
        headers["Cookie"] = cookies
    url = "https://exhentai.org" if cookies else "https://e-hentai.org"
    params = {"f_search": title}
    async with aiohttp.ClientSession(headers=headers) as session:
        resp = await session.get(url, proxy=proxy, params=params)
        if res := EHentaiResponseAioHttp(await resp.text(), str(resp.url)):
            if not res.raw:
                # 如果第一次没找到，使搜索结果包含被删除的部分，并重新搜索
                params["advsearch"] = 1
                params["f_sname"] = "on"
                params["f_stags"] = "on"
                params["f_sh"] = "on"
                resp = await session.get(url, proxy=proxy, params=params)
                res = EHentaiResponseAioHttp(await resp.text(), str(resp.url))
            return await search_result_filter(res, proxy, hide_img)
        return ["EHentai 暂时无法使用"]


async def search_result_filter(
    res: Union[EHentaiResponse, EHentaiResponseAioHttp],
    proxy: Optional[str],
    hide_img: bool,
) -> List[str]:
    if not res.raw:
        return ["EHentai 搜索结果为空"]
        # 尽可能过滤掉图集，除非只有图集
    if non_image_set_list := [i for i in res.raw if i.type != "Image Set"]:
        res.raw = non_image_set_list
    selected_res = res.raw[0]
    # 优先找汉化版
    for i in res.raw:
        if "translated" in i.tags and "chinese" in i.tags:
            selected_res = i
            break
    thumbnail = await handle_img(
        selected_res.thumbnail, proxy, hide_img, config.exhentai_cookies
    )
    res_list = [
        "EHentai 搜索结果",
        thumbnail,
        selected_res.title,
        f"类型：{selected_res.type}",
        f"日期：{selected_res.date}",
        f"来源：{selected_res.url}",
        "-" * 20,
        f"搜索页面：{res.url}",
    ]
    return ["\n".join([i for i in res_list if i != ""])]



from typing import List, Optional

from PicImageSearch import Ascii2D, Network
from PicImageSearch.model import Ascii2DResponse

from .utils import handle_img, shorten_url


async def ascii2d_search(url: str, proxy: Optional[str], hide_img: bool) -> List[str]:
    async with Network(proxies=proxy) as client:
        ascii2d_color = Ascii2D(client=client)
        ascii2d_bovw = Ascii2D(bovw=True, client=client)
        color_res = await ascii2d_color.search(url)
        bovw_res = await ascii2d_bovw.search(url)

        async def get_final_res(res: Ascii2DResponse) -> str:
            if not res or not res.raw:
                return ""
            if not res.raw[0].url:
                res.raw[0] = res.raw[1]
            thumbnail = await handle_img(res.raw[0].thumbnail, proxy, hide_img)
            _url = shorten_url(res.raw[0].url) if res.raw[0] else ""
            res_list = [
                f"{thumbnail}",
                f"{res.raw[0].title}" if res.raw[0].title else "",
                f"Author：{res.raw[0].author}" if res.raw[0].author else "",
                _url,
            ]
            return "\n".join([i for i in res_list if i != ""])

        color_final_res = await get_final_res(color_res)
        bovw_final_res = await get_final_res(bovw_res)
        if color_final_res == bovw_final_res:
            if color_final_res == "":
                return ["Ascii2D 暂时无法使用"]
            return [f"Ascii2D 色合検索与特徴検索結果完全一致\n{color_final_res}"]

        return [
            f"Ascii2D 色合検索結果\n{color_final_res}",
            f"Ascii2D 特徴検索結果\n{bovw_final_res}",
        ]



import re
from typing import List, Optional

from PicImageSearch import Network, SauceNAO

from .ascii2d import ascii2d_search
from .config import config
from .ehentai import ehentai_title_search
from .utils import get_source, handle_img, shorten_url
from .whatanime import whatanime_search


async def saucenao_search(
    url: str, mode: str, proxy: Optional[str], hide_img: bool
) -> List[str]:
    saucenao_db = {
        "all": 999,
        "pixiv": 5,
        "danbooru": 9,
        "anime": 21,
        "doujin": [18, 38],
    }
    async with Network(proxies=proxy) as client:
        if isinstance(db := saucenao_db[mode], list):
            saucenao = SauceNAO(client=client, api_key=config.saucenao_api_key, dbs=db)
        else:
            saucenao = SauceNAO(client=client, api_key=config.saucenao_api_key, db=db)
        res = await saucenao.search(url)
        final_res = []
        if res and res.raw:
            selected_res = res.raw[0]
            ext_urls = selected_res.origin["data"].get("ext_urls")
            # 如果结果为 pixiv ，尝试找到原始投稿，避免返回盗图者的投稿
            if selected_res.index_id == saucenao_db["pixiv"]:
                pixiv_res_list = list(
                    filter(
                        lambda x: x.index_id == saucenao_db["pixiv"]
                        and x.url
                        and abs(x.similarity - selected_res.similarity) < 5,
                        res.raw,
                    )
                )
                if len(pixiv_res_list) > 1:
                    selected_res = min(
                        pixiv_res_list,
                        key=lambda x: int(re.search(r"\d+", x.url).group()),  # type: ignore
                    )
            # 如果地址有多个，优先取 danbooru
            elif ext_urls and len(ext_urls) > 1:
                for i in ext_urls:
                    if "danbooru" in i:
                        selected_res.url = i
            thumbnail = await handle_img(selected_res.thumbnail, proxy, hide_img)
            if selected_res.origin["data"].get("source"):
                source = shorten_url(selected_res.origin["data"]["source"])
            else:
                source = shorten_url(await get_source(selected_res.url, proxy))
            # 如果结果为 doujin ，尝试返回日文标题而不是英文标题
            if selected_res.index_id == saucenao_db["doujin"] and selected_res.origin[
                "data"
            ].get("jp_name"):
                selected_res.title = selected_res.origin["data"]["jp_name"]
            _url = shorten_url(selected_res.url)
            res_list = [
                f"SauceNAO（{selected_res.similarity}%）",
                f"{thumbnail}",
                f"{selected_res.title}",
                f"Author：{selected_res.author}" if selected_res.author else "",
                _url,
                f"Source：{source}" if source else "",
            ]
            final_res.append("\n".join([i for i in res_list if i != ""]))
            if (
                selected_res.similarity < config.saucenao_low_acc
                and mode == "anime"
                or selected_res.similarity >= config.saucenao_low_acc
                and "anidb.net" in _url
            ):
                final_res.extend(await whatanime_search(url, proxy, hide_img))
            elif (
                selected_res.similarity >= config.saucenao_low_acc and mode == "doujin"
            ):
                final_res.extend(
                    await ehentai_title_search(selected_res.title, proxy, hide_img)
                )
            elif (
                selected_res.similarity < config.saucenao_low_acc
                and config.use_ascii2d_when_low_acc
            ):
                final_res.append(f"相似度 {selected_res.similarity}% 过低，自动使用 Ascii2D 进行搜索")
                final_res.extend(await ascii2d_search(url, proxy, hide_img))
        else:
            final_res.append("SauceNAO 暂时无法使用，自动使用 Ascii2D 进行搜索")
            final_res.extend(await ascii2d_search(url, proxy, hide_img))
        return final_res



import base64
import re
from typing import Optional

import aiohttp
from pyquery import PyQuery
from yarl import URL


# 将图片转化为 base64
async def get_pic_base64_by_url(
    url: str, proxy: Optional[str], cookies: Optional[str] = None
) -> str:
    headers = {}
    if cookies:
        headers["Cookie"] = cookies
    async with aiohttp.ClientSession(headers=headers) as session:
        async with session.get(url, proxy=proxy) as resp:
            if resp.status == 200:
                return base64.b64encode(await resp.read()).decode()
    return ""


async def handle_img(
    url: str, proxy: Optional[str], hide_img: bool, cookies: Optional[str] = None
) -> str:
    if hide_img:
        return ""
    img_base64 = await get_pic_base64_by_url(url, proxy, cookies)
    if img_base64:
        return f"[CQ:image,file=base64://{img_base64}]"
    return f"图片下载失败: {url}"


async def get_source(url: str, proxy: Optional[str]) -> str:
    source = ""
    async with aiohttp.ClientSession() as session:
        if URL(url).host == "danbooru.donmai.us":
            async with session.get(url, proxy=proxy) as resp:
                if resp.status == 200:
                    html = await resp.text()
                    source = PyQuery(html)(".image-container").attr(
                        "data-normalized-source"
                    )
        elif URL(url).host in ["yande.re", "konachan.com"]:
            async with session.get(url, proxy=proxy) as resp:
                if resp.status == 200:
                    html = await resp.text()
                    source = PyQuery(html)("#stats li:contains(Source) a").attr("href")
        elif URL(url).host == "gelbooru.com":
            async with session.get(url, proxy=proxy) as resp:
                if resp.status == 200:
                    html = await resp.text()
                    source = PyQuery(html)("#tag-list li:contains(Source) a").attr(
                        "href"
                    )
    return str(source)


def shorten_url(url: str) -> str:
    pid_search = re.compile(
        r"(?:pixiv.+(?:illust_id=|artworks/)|/img-original/img/(?:\d+/){6})(\d+)"
    )
    if pid_search.search(url):
        return f"https://pixiv.net/i/{pid_search.search(url)[1]}"  # type: ignore
    if URL(url).host == "danbooru.donmai.us":
        return url.replace("/post/show/", "/posts/")
    return url



from typing import Any, Dict, List, Optional


class Result:
    def __init__(self, mapping: Dict[str, Any]):
        self.ascii2d: List[str] = []
        self.saucenao: List[str] = []
        self.ex: List[str] = []
        self.image_md5: Optional[str] = None
        self.update_at: Optional[str] = None
        self.mode: str = "all"
        for key, value in mapping.items():
            setattr(self, key, value)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['YetAnotherPicSearch']

package_data = \
{'': ['*']}

install_requires = \
['PicImageSearch>=3.2.0,<4.0.0',
 'aiohttp>=3.8.1,<4.0.0',
 'arrow>=1.2.0,<2.0.0',
 'lxml>=4.8.0,<5.0.0',
 'nonebot-adapter-onebot>=2.0.0b1,<3.0.0',
 'nonebot2>=2.0.0b1,<3.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'pyquery>=1.4.3,<2.0.0',
 'tinydb>=4.6.1,<5.0.0',
 'yarl>=1.7.2,<2.0.0']

setup_kwargs = {
    'name': 'yetanotherpicsearch',
    'version': '1.4.1',
    'description': 'Yet Another Picture Search Nonebot Plugin',
    'long_description': '# YetAnotherPicSearch\n\n基于 [nonebot/nonebot2](https://github.com/nonebot/nonebot2) 及 [kitUIN/PicImageSearch](https://github.com/kitUIN/PicImageSearch) 的另一个 Nonebot 搜图插件。\n\n目前适配的是 `OneBot V11` ，没适配 QQ 频道。\n\n主要受到 [Tsuk1ko/cq-picsearcher-bot](https://github.com/Tsuk1ko/cq-picsearcher-bot) 的启发。 而我只需要基础的搜图功能，于是忍不住自己也写了一个。\n\n目前支持的搜图服务：\n\n- [saucenao](https://saucenao.com)\n- [whatanime](https://trace.moe)\n- [ascii2d](https://ascii2d.net)\n- [exhentai](https://exhentai.org)\n\n详细说明请移步：[docs](docs)\n\n## 感谢以下项目或服务（不分先后）\n\n- [saucenao](https://saucenao.com)\n- [ascii2d](https://ascii2d.net)\n- [trace.moe](https://trace.moe) ([GitHub](https://github.com/soruly/trace.moe))\n- [exhentai](https://exhentai.org)\n- [cq-picsearcher-bot](https://github.com/Tsuk1ko/cq-picsearcher-bot)\n- [PicImageSearch](https://github.com/kitUIN/PicImageSearch)\n',
    'author': 'NekoAria',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/NekoAria/YetAnotherPicSearch',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
